#include "puzzle.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
  int x;
  int y;
  int z;
} pos_t;

int busy[3][3][3];

int *check_constraints(int cube, int x, int y, int z, int *puzzle, pos_t *p)
{
  int c1, c2, c3, c4;
  int delta;
  int delta_x1, delta_x2;
  int delta_y1, delta_y2;
  int delta_z1, delta_z2;
  int aligned;
  int d1, d2;
  int u1, u2, u3;
  int v1, v2, v3;
  int vec1, vec2, vec3;

  while (*puzzle != CUBE && *puzzle != -1) {
    switch (*puzzle) {
    case TOUCH:
      c1 = puzzle[1];
      delta = abs(p[cube].x - p[c1].x)
            + abs(p[cube].y - p[c1].y)
            + abs(p[cube].z - p[c1].z);
printf("touch p[%d] (%d %d %d) p[%d] (%d %d %d) delta %d\n", cube, p[cube].x, p[cube].y, p[cube].z, c1, p[c1].x, p[c1].y, p[c1].z, delta);
      if (delta != 1) return NULL;
      puzzle += 2;
      break;
    case ALIGNED:
    case NOT_ALIGNED:
      c1 = puzzle[1];
      c2 = puzzle[2];
      delta_x1 = p[c2].x - p[c1].x;
      delta_x2 = p[cube].x - p[c2].x;
      delta_y1 = p[c2].y - p[c1].y;
      delta_y2 = p[cube].y - p[c2].y;
      delta_z1 = p[c2].z - p[c1].z;
      delta_z2 = p[cube].z - p[c2].z;
      aligned = delta_x1 == delta_x2
              && delta_y1 == delta_y2
              && delta_z1 == delta_z2;
printf("%s cube %d c1c2 %d %d aligned %d\n", *puzzle == ALIGNED ? "ALIGNED" : "NOT_ALIGNED", cube, c1, c2, aligned);
      if (*puzzle == ALIGNED && !aligned) return NULL;
      if (*puzzle == NOT_ALIGNED && aligned) return NULL;
      puzzle += 3;
      break;
    case ORDER:
      c1 = puzzle[1];
      c2 = puzzle[2];
      d1 = p[c1].x * 9 + p[c1].y * 3 + p[c1].z;
      d2 = p[c2].x * 9 + p[c2].y * 3 + p[c2].z;
printf("order p[%d] (%d %d %d) p[%d] (%d %d %d) d1 %d d2 %d\n", c1, p[c1].x, p[c1].y, p[c1].z, c2, p[c2].x, p[c2].y, p[c2].z, d1, d2);
      if (d2 >= d1) return NULL;
      puzzle += 3;
      break;
    case LEFT:
      c1 = puzzle[1];
      c2 = puzzle[2];
      c3 = puzzle[3];
      c4 = puzzle[4];
      /* compute vectorial product */
      u1 = p[c2].x - p[c1].x; u2 = p[c2].y - p[c1].y; u3 = p[c2].z - p[c1].z;
      v1 = p[c3].x - p[c1].x; v2 = p[c3].y - p[c1].y; v3 = p[c3].z - p[c1].z;
      vec1 = u2 * v3 - u3 * v2;
      vec2 = u3 * v1 - u1 * v3;
      vec3 = u1 * v2 - u2 * v1;
      /* putting vec at c4 its end should be at cube */
      if (   p[c4].x + vec1 != p[cube].x
          || p[c4].y + vec2 != p[cube].y
          || p[c4].z + vec3 != p[cube].z)
        return NULL;
      puzzle += 5;
      break;
    case VEC:
      c1 = puzzle[1];
      c2 = puzzle[2];
      c3 = puzzle[3];
      vec1 = p[c2].x - p[c1].x;
      vec2 = p[c2].y - p[c1].y;
      vec3 = p[c2].z - p[c1].z;
printf("vec p[%d] (%d %d %d) p[%d] (%d %d %d) p[%d] (%d %d %d) vec(%d %d %d) cube [%d] (%d %d %d) c3+vec (%d %d %d)\n",
       c1, p[c1].x, p[c1].y, p[c1].z, c2, p[c2].x, p[c2].y, p[c2].z, c3, p[c3].x, p[c3].y, p[c3].z, vec1, vec2, vec3, cube, p[cube].x, p[cube].y, p[cube].z,
       p[c3].x + vec1, p[c3].y + vec2, p[c3].z + vec3);
      /* putting vec at c3 its end should be at cube */
      if (   p[c3].x + vec1 != p[cube].x
          || p[c3].y + vec2 != p[cube].y
          || p[c3].z + vec3 != p[cube].z)
        return NULL;
      puzzle += 4;
      break;
    }
  }
  return puzzle;
}

void solve(int cube, pos_t *p, int *puzzle, char *piece)
{
  int x, y, z;
  int *pz;
  if (puzzle[0] == -1) {
    int x, y, z;
    int v = 0;
    /* get the solution as a value in [0 .. 2^27-1] */
    for (x = 0; x < 3; x++) for (y = 0; y < 3; y++) for (z = 0; z < 3; z++) {
      v <<= 1;
      if (busy[x][y][z] != -1) v |= 1;
    }
    printf("solution %s 0x%x\n", piece, v);
    return;
  }
  if (puzzle[0] != CUBE) abort();
  if (puzzle[1] != cube) abort();
  puzzle += 2;
  for (x = 0; x < 3; x++) for (y = 0; y < 3; y++) for (z = 0; z < 3; z++) {
    if (busy[x][y][z] != -1) continue;
#if 0
/* force one cube to be in a corner of the big cube */
if (cube == 0) {
  int X = x-1;
  int Y = y-1;
  int Z = z-1;
  if (X*X+Y*Y+Z*Z != 3) continue;
}
#endif
    p[cube].x = x;
    p[cube].y = y;
    p[cube].z = z;
    pz = check_constraints(cube, x, y, z, puzzle, p);
    if (pz == NULL) continue;
    busy[x][y][z] = cube;
    solve(cube + 1, p, pz, piece);
    busy[x][y][z] = -1;
  }
}

void go(int *puzzle, char *piece)
{
  pos_t p[27];
  int x, y, z;

  for (x = 0; x < 3; x++) for (y = 0; y < 3; y++) for (z = 0; z < 3; z++)
    busy[x][y][z] = -1;

  solve(0, p, puzzle, piece);
}

/*
 * 0
 * 1 2
 */
int *piece_1_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1, //ORDER, 0, 2,
  -1
};
int *piece_1_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1, ORDER, 0, 2,
  -1
};

/*
 * 0
 * 1 2 3
 */
int *piece_2_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, ALIGNED, 1, 2,
  -1
};

/*
 * 0 1 2
 *   3
 */
int *piece_3_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, ALIGNED, 0, 1, //ORDER, 0, 2,
  CUBE, 3,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  -1
};
int *piece_3_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, ALIGNED, 0, 1, ORDER, 0, 2,
  CUBE, 3,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  -1
};

/*
 *   2 3
 * 0 1
 */
int *piece_4_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, VEC, 0, 1, 2, //ORDER, 0, 3,
  -1
};
int *piece_4_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, VEC, 0, 1, 2, ORDER, 0, 3,
  -1
};

/*
 *    3
 *   2
 * 0 1
 */
int *piece_5_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, LEFT, 1, 2, 0, 2, //ORDER, 0, 3,
  -1
};
int *piece_5_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, LEFT, 1, 2, 0, 2, ORDER, 0, 3,
  -1
};

/*
 * 3
 *  1 2
 *  0
 */
int *piece_6_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1, //ORDER, 0, 2,
  CUBE, 3,  TOUCH, 1, LEFT, 1, 0, 2, 1, //ORDER, 0, 3,
  -1
};
int *piece_6_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1, ORDER, 0, 2,
  CUBE, 3,  TOUCH, 1, LEFT, 1, 0, 2, 1, ORDER, 0, 3,
  -1
};

/*
 *    3
 * 1 2
 * 0
 */
int *piece_7_full = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, LEFT, 1, 0, 2, 2, //ORDER, 0, 3,
  -1
};
int *piece_7_constrained = (int []) {
  CUBE, 0,
  CUBE, 1,  TOUCH, 0,
  CUBE, 2,  TOUCH, 1, NOT_ALIGNED, 0, 1,
  CUBE, 3,  TOUCH, 2, LEFT, 1, 0, 2, 2, ORDER, 0, 3,
  -1
};

int main(void)
{
  go(piece_1_full, "full piece 1");
  go(piece_1_constrained, "constrained piece 1");
  go(piece_2_full, "full piece 2");
  go(piece_3_full, "full piece 3");
  go(piece_3_constrained, "constrained piece 3");
  go(piece_4_full, "full piece 4");
  go(piece_4_constrained, "constrained piece 4");
  go(piece_5_full, "full piece 5");
  go(piece_5_constrained, "constrained piece 5");
  go(piece_6_full, "full piece 6");
  go(piece_6_constrained, "constrained piece 6");
  go(piece_7_full, "full piece 7");
  go(piece_7_constrained, "constrained piece 7");
  return 0;
}
